class cached_property:

    """Cache property value on first access.

    >>> value = object()
    >>>
    >>> class C:
    ...    @cached_property
    ...    def x(self):
    ...        return value
    ...
    >>> isinstance(C.x, cached_property)
    True
    >>> c = C()
    >>> 'x' in c.__dict__
    False
    >>> c.x is value
    True
    >>> 'x' in c.__dict__
    True

    """

    def __init__(self, fget):
        self.fget = fget
        self.__name__ = fget.__name__
        self.__doc__ = fget.__doc__

    def __get__(self, obj, cls=None):
        if obj is None:  # Property accessed via class
            return self
        obj.__dict__[self.__name__] = self.fget(obj)
        return obj.__dict__[self.__name__]



import os
from contextlib import contextmanager
from doctest import DocTestSuite
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest import TestCase

from runcommands.commands import copy_file


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite('runcommands.util.decorators'))
    tests.addTests(DocTestSuite('runcommands.util.misc'))
    tests.addTests(DocTestSuite('runcommands.util.path'))
    tests.addTests(DocTestSuite('runcommands.util.string'))
    return tests


@contextmanager
def copy(contents, destination_dir=False, **kwargs):
    with NamedTemporaryFile('w', delete=False) as tp:
        tp.write(contents)

    source = tp.name

    if destination_dir:
        with TemporaryDirectory() as destination:
            path = copy_file(source, destination, **kwargs)
            yield source, destination, path
            os.remove(source)
    else:
        destination = source + '.copy'
        path = copy_file(source, destination, **kwargs)
        yield source, destination, path
        os.remove(source)
        os.remove(path)


class TestCopyFileCommand(TestCase):

    def assertDestination(self, source, destination, path, expected_contents):
        self.assertExists(path)
        self.assertNotEqual(source, destination)
        self.assertNotEqual(source, path)
        self.assertContentsEqual(path, expected_contents)
        if not os.path.isdir(destination):
            self.assertTrue(path.endswith('.copy'))

    def assertExists(self, path):
        self.assertTrue(os.path.isfile(path))

    def assertContentsEqual(self, path, expected_contents):
        with open(path) as fp:
            read_contents = fp.read()
        self.assertEqual(read_contents, expected_contents)

    def test_copy(self):
        with copy('xyz') as (source, destination, path):
            self.assertDestination(source, destination, path, 'xyz')

    def test_copy_template_format(self):
        with copy('{xyz} ${{xyz}}', template=True, context={'xyz': 123}) as paths:
            source, destination, path = paths
            self.assertDestination(source, destination, path, '123 ${xyz}')

    def test_copy_template_string(self):
        with copy('${xyz} $${xyz}', template='string', context={'xyz': 123}) as paths:
            source, destination, path = paths
            self.assertDestination(source, destination, path, '123 ${xyz}')

    def test_copy_to_directory(self):
        with copy('xyz', destination_dir=True) as paths:
            source, destination, path = paths
            self.assertDestination(source, destination, path, 'xyz')
            self.assertEqual(os.path.dirname(path), destination)
            self.assertNotEqual(os.path.dirname(source), os.path.dirname(path))
            self.assertEqual(os.path.basename(source), os.path.basename(path))



#!/usr/bin/env python3
import getpass
import importlib
import os
import shutil
import sys
import unittest


if os.path.abspath(sys.argv[0]) == os.path.abspath(__file__):
    # When running this module directly via `./commands.py` or
    # `python commands.py`:
    #
    # - Ensure virtual env is activated and minimal
    # - Ensure minimal set of dependencies are installed
    # - Ensure runcommands project directory is first in sys.path

    def check_dependency(name, dist_name=None, action=None):
        if sys.version_info > (3, 5):
            exc_type = ModuleNotFoundError  # noqa
        else:
            exc_type = ImportError
        try:
            importlib.import_module(name)
        except exc_type:
            if action:
                action(name)
            else:
                dist_name = dist_name or name
                sys.stderr.write('Run `pip install {dist_name}` first\n'.format_map(locals()))
                sys.exit(2)

    virtual_env = os.getenv('VIRTUAL_ENV')
    if not virtual_env:
        sys.stderr.write('No virtual env active\n')
        sys.stderr.write('Run `python -m venv .venv` first, then activate the virtual env\n')
        sys.exit(1)

    check_dependency('jinja2')
    check_dependency('yaml', 'pyyaml')
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))


from runcommands import command  # noqa: E402
from runcommands.args import arg  # noqa: E402
from runcommands.commands import copy_file as _copy_file, local  # noqa: E402
from runcommands.commands import git_version, release  # noqa: E402,F401
from runcommands.commands.release import get_current_branch, get_latest_tag  # noqa: E402
from runcommands.util import abort, asset_path, confirm, printer  # noqa: E402


@command
def virtualenv(where='.venv', python='python', overwrite=False):
    exists = os.path.exists(where)

    def create():
        local((python, '-m', 'venv', where))
        printer.success(
            'Virtualenv created; activate it by running `source {where}/bin/activate`'
            .format_map(locals()))

    if exists:
        if overwrite:
            printer.warning('Overwriting virtualenv', where, 'with', python)
            shutil.rmtree(where)
            create()
        else:
            printer.info('Virtualenv', where, 'exists; pass --overwrite to re-create it')
    else:
        printer.info('Creating virtualenv', where, 'with', python)
        create()


@command
def install(where='.venv', python='python', upgrade=False, overwrite=False):
    virtualenv(where=where, python=python, overwrite=overwrite)
    pip = '{where}/bin/pip'.format(where=where)
    local((
        pip, 'install',
        ('--upgrade', '--upgrade-strategy', 'eager') if upgrade else None,
        '--editable', '.[dev]',
        ('pip', 'setuptools') if upgrade else None,
    ), echo=True)


@command
def install_completion(
    shell: arg(
        choices=('bash', 'fish'),
        help='Shell to install completion for',
    ),
    to: arg(
        help='~/.bashrc.d/runcommands.rc or ~/.config/fish/runcommands.fish',
    ) = None,
    base_command: arg(
        help='Dotted path to base command',
    ) = None,
    base_command_name: arg(
        short_option='-B',
        help='Name of base command (if different from implementation name)'
    ) = None,
    overwrite: arg(
        help='Overwrite if exists',
    ) = False,
):
    """Install command line completion script.

    Currently, bash and fish are supported. The corresponding script
    will be copied to an appropriate directory. If the script already
    exists at that location, it will be overwritten by default.

    """
    if base_command:
        if not base_command_name:
            _, base_command_name = base_command.rsplit('.', 1)
        source_base_name = 'runcommands-base-command'
        to_file_name = base_command_name
        template_type = 'string'
        template_context = {
            'base_command_path': base_command,
            'base_command_name': base_command_name,
        }
    else:
        source_base_name = 'runcommands'
        to_file_name = ''
        template_type = None
        template_context = {}

    if shell == 'bash':
        ext = 'rc'
        to = to or '~/.bashrc.d'
    elif shell == 'fish':
        ext = 'fish'
        to = to or '~/.config/fish'

    if base_command:
        to = '{to}/{base_command_name}.{ext}'.format_map(locals())

    source_path = 'runcommands:completion/{shell}/{source_base_name}.{ext}'
    source_path = source_path.format_map(locals())
    source = asset_path(source_path)

    destination = os.path.expanduser(to)

    if os.path.isdir(destination):
        destination = os.path.join(destination, os.path.basename(source))

    printer.info('Installing', shell, 'completion script to:\n    ', destination)

    if os.path.exists(destination):
        if overwrite:
            printer.info('Overwriting:\n    {destination}'.format_map(locals()))
        else:
            message = 'File exists. Overwrite?'.format_map(locals())
            overwrite = confirm(message, abort_on_unconfirmed=True)

    _copy_file(source, destination, template=template_type, context=template_context)
    printer.info('Installed; remember to:\n    source {destination}'.format_map(locals()))


@command
def test(*tests, fail_fast=False, with_coverage=True, with_lint=True):
    original_working_directory = os.getcwd()

    if tests:
        num_tests = len(tests)
        s = '' if num_tests == 1 else 's'
        printer.header('Running {num_tests} test{s}...'.format_map(locals()))
    else:
        coverage_message = ' with coverage' if with_coverage else ''
        printer.header('Running tests{coverage_message}...'.format_map(locals()))

    runner = unittest.TextTestRunner(failfast=fail_fast)
    loader = unittest.TestLoader()

    if with_coverage:
        from coverage import Coverage
        coverage = Coverage(source=['runcommands'])
        coverage.start()

    if tests:
        runner.run(loader.loadTestsFromNames(tests))
    else:
        tests = loader.discover('.')
        result = runner.run(tests)
        if not result.errors:
            if with_coverage:
                coverage.stop()
                coverage.report()
            if with_lint:
                printer.header('Checking for lint...')
                # XXX: The test runner apparently changes CWD.
                os.chdir(original_working_directory)
                lint()


@command
def tox(envs: 'Pass -e option to tox with the specified environments' = (),
        recreate: 'Pass --recreate flag to tox' = False,
        clean: 'Remove tox directory first' = False):
    if clean:
        local('rm -rf .tox', echo=True)
    local((
        'tox',
        ('-e', ','.join(envs)) if envs else None,
        '--recreate' if recreate else None,
    ))


@command
def lint(show_errors: arg(help='Show errors') = True,
         disable_ignore: arg(no_inverse=True, help='Don\'t ignore any errors') = False,
         disable_noqa: arg(no_inverse=True, help='Ignore noqa directives') = False):
    result = local((
        'flake8', '.',
        '--ignore=' if disable_ignore else None,
        '--disable-noqa' if disable_noqa else None,
    ), stdout='capture', raise_on_error=False)
    pieces_of_lint = len(result.stdout_lines)
    if pieces_of_lint:
        ess = '' if pieces_of_lint == 1 else 's'
        colon = ':' if show_errors else ''
        message = ['{pieces_of_lint} piece{ess} of lint found{colon}'.format_map(locals())]
        if show_errors:
            message.append(result.stdout.rstrip())
        message = '\n'.join(message)
        abort(1, message)
    else:
        printer.success('No lint found')


@command
def clean(verbose=False):
    """Clean up.

    Removes:

        - ./build/
        - ./dist/
        - **/__pycache__
        - **/*.py[co]

    Skips hidden directories.

    """
    root = os.getcwd()

    rmdir('build', verbose)
    rmdir('dist', verbose)

    for path, dirs, files in os.walk(root):
        rel_path = os.path.relpath(path, root)

        if rel_path == '.':
            rel_path = ''

        if rel_path.startswith('.'):
            continue

        for d in dirs:
            if d == '__pycache__':
                rmdir(os.path.join(rel_path, d))

        for f in files:
            if f.endswith('.pyc') or f.endswith('.pyo'):
                rmfile(os.path.join(rel_path, f), verbose)


@command
def build_docs(source='docs', destination='docs/_build', builder='html', clean=False):
    if clean:
        printer.info('Removing {destination}...'.format_map(locals()))
        shutil.rmtree(destination)
    local((
        'sphinx-build',
        '-b',
        builder,
        source,
        destination,
    ))


@command
def make_dist(
    version: arg(help='Tag/version to release [latest tag]') = None,
    quiet=False
):
    """Make a distribution for upload to PyPI.

    Switches to the specified tag or branch, makes the distribution,
    then switches back to the original branch.

    Intended to be run from the develop branch. If a tag is already
    checked out, the develop branch will be checked out first and then
    switched back to after making the distribution.

    """
    current_branch = get_current_branch()
    original_branch = 'develop' if current_branch == 'HEAD' else current_branch
    version = version or get_latest_tag()
    stdout = 'hide' if quiet else None

    printer.header('Making dist for {version}'.format_map(locals()))

    if version != current_branch:
        if current_branch == 'HEAD':
            printer.warning('Not on a branch; checking out develop first')
        else:
            printer.info('Currently on branch', current_branch)
        printer.info('Checking out', version)
        # XXX: Hide warning about detached HEAD state
        result = local(('git', 'checkout', version), stdout=stdout, stderr='capture')
        if result.failed:
            print(result.stderr, file=sys.stderr)

    printer.info('Removing dist directory')
    rmdir('dist', verbose=not quiet)

    printer.info('Making sdist for', version)
    local('python setup.py sdist', stdout=stdout)

    if version != current_branch:
        printer.info('Switching back to', original_branch)
        # XXX: Hide message about previous HEAD state and branch info
        result = local(('git', 'checkout', original_branch), stdout='hide', stderr='capture')
        if result.failed:
            print(result.stderr, file=sys.stderr)


@command
def upload_dists(
    make: arg(help='Make dist first? [yes]') = True,
    version: arg(help='Version/tag to release [latest tag]') = None,
    quiet: arg(help='Make dist quietly? [no]') = False,
    username: arg(help='Twine username [$USER]') = None,
    password_command: arg(
        help='Command to retrieve twine password '
             '(e.g. `password-manager show-password PyPI`) '
             '[twine prompt]'
    ) = None
):
    """Upload distributions in ./dist using ``twine``."""
    if make:
        printer.header('Making and uploading distributions')
        make_dist(quiet)
    else:
        printer.header('Uploading distributions')

    dists = os.listdir('dist')
    if not dists:
        abort(1, 'No distributions found in dist directory')

    paths = [os.path.join('dist', file) for file in dists]

    printer.info('Found distributions:')
    for path in paths:
        printer.info('  -', path)

    if not confirm('Continue?'):
        abort()

    if not username:
        username = getpass.getuser()
    environ = {'TWINE_USERNAME': username}

    if password_command:
        printer.info('Retrieving password via `{password_command}`...'.format_map(locals()))
        result = local(password_command, stdout='capture')
        password = result.stdout.strip()
        environ['TWINE_PASSWORD'] = password

    printer.warning('TWINE_USERNAME:', username)
    if password:
        printer.warning('TWINE_PASSWORD:', '*' * len(password))

    for path in paths:
        if confirm('Upload dist?: {path}'.format_map(locals())):
            local(('twine', 'upload', path), environ=environ)
        else:
            printer.warning('Skipped dist:', path)


# Utilities


def rmfile(name, verbose=False):
    if os.path.isfile(name):
        os.remove(name)
        if verbose:
            printer.info('Removed file:', name)
    else:
        if verbose:
            printer.info('File not present:', name)


def rmdir(name, verbose=False):
    if os.path.isdir(name):
        shutil.rmtree(name)
        if verbose:
            printer.info('Removed directory:', name)
    else:
        if verbose:
            printer.info('Directory not present:', name)


if __name__ == '__main__':
    from runcommands.__main__ import main
    sys.exit(main())



import shlex
import sys

from ..args import arg
from ..command import command
from ..result import Result
from ..util import flatten_args, isatty, StreamOptions
from .local import local


@command
def remote(cmd: arg(container=list),
           host,
           user=None,
           port=None,
           sudo=False,
           run_as=None,
           shell='/bin/sh',
           cd=None,
           environ: arg(container=dict) = None,
           paths=(),
           # Args passed through to local command:
           stdout: arg(type=StreamOptions) = None,
           stderr: arg(type=StreamOptions) = None,
           echo=False,
           raise_on_error=True,
           dry_run=False,
           ) -> Result:
    """Run a remote command via SSH.

    Runs a remote shell command using ``ssh`` in a subprocess like so::

        ssh -q [-t] [<user>@]<host> [sudo [-u <run_as>] -H] /bin/sh -c '
            [cd <cd> &&]
            [export XYZ="xyz" &&]
            [export PATH="<path>" &&]
            <cmd>
        '

    Args:
        cmd (list|str): The command to run. If this is a list, it will
            be flattened into a string.
        host (str): Remote host to SSH into.
        user (str): Remote user to log in as (defaults to current local
            user).
        port (int): SSH port on remote host.
        sudo (bool): Run the remote command as root using ``sudo``.
        run_as (str): Run the remote command as a different user using
            ``sudo -u <run_as>``.
        shell (str): The remote user's default shell will be used to run
            the remote command unless this is set to a different shell.
        cd (str): Where to run the command on the remote host.
        environ (dict): Extra environment variables to set on the remote
            host.
        paths (list): Additional paths to prepend to the remote
            ``$PATH``.
        stdout: See :obj:`runcommands.commands.local`.
        stderr: See :obj:`runcommands.commands.local`.
        echo: See :obj:`runcommands.commands.local`.
        raise_on_error: See :obj:`runcommands.commands.local`.
        dry_run: See :obj:`runcommands.commands.local`.

    """
    if not isinstance(cmd, str):
        cmd = flatten_args(cmd, join=True)

    ssh_options = ['-q']
    if isatty(sys.stdin):
        ssh_options.append('-t')
    if port is not None:
        ssh_options.extend(('-p', port))

    ssh_connection_str = '{user}@{host}'.format_map(locals()) if user else host

    remote_cmd = []

    if sudo:
        remote_cmd.extend(('sudo', '-H'))
    elif run_as:
        remote_cmd.extend(('sudo', '-H', '-u', run_as))

    remote_cmd.extend((shell, '-c'))

    inner_cmd = []

    if cd:
        inner_cmd.append('cd {cd}'.format_map(locals()))

    if environ:
        inner_cmd.extend('export {k}="{v}"'.format_map(locals()) for k, v in environ.items())

    if paths:
        inner_cmd.append('export PATH="{path}:$PATH"'.format(path=':'.join(paths)))

    inner_cmd.append(cmd)
    inner_cmd = ' &&\n    '.join(inner_cmd)
    inner_cmd = '\n    {inner_cmd}\n'.format_map(locals())
    inner_cmd = shlex.quote(inner_cmd)

    remote_cmd.append(inner_cmd)
    remote_cmd = ' '.join(remote_cmd)

    args = ('ssh', ssh_options, ssh_connection_str, remote_cmd)
    return local(
        args, stdout=stdout, stderr=stderr, echo=echo, raise_on_error=raise_on_error,
        dry_run=dry_run)



import os
import shutil
import string
import tempfile

from ..args import arg, bool_or
from ..command import command


@command
def copy_file(source, destination, follow_symlinks=True,
              template: arg(type=bool_or(str), choices=('format', 'string')) = False,
              context=None):
    """Copy source file to destination.

    The destination may be a file path or a directory. When it's a
    directory, the source file will be copied into the directory
    using the file's base name.

    When the source file is a template, ``context`` will be used as the
    template context. The supported template types are 'format' and
    'string'. The former uses ``str.format_map()`` and the latter uses
    ``string.Template()``.

    .. note:: :func:`shutil.copy()` from the standard library is used to
        do the copy operation.

    """
    if not template:
        # Fast path for non-templates.
        return shutil.copy(source, destination, follow_symlinks=follow_symlinks)

    if os.path.isdir(destination):
        destination = os.path.join(destination, os.path.basename(source))

    with open(source) as source:
        contents = source.read()

    if template is True or template == 'format':
        contents = contents.format_map(context)
    elif template == 'string':
        string_template = string.Template(contents)
        contents = string_template.substitute(context)
    else:
        raise ValueError('Unknown template type: %s' % template)

    with tempfile.NamedTemporaryFile('w', delete=False) as temp_file:
        temp_file.write(contents)

    path = shutil.copy(temp_file.name, destination)
    os.remove(temp_file.name)
    return path



import os
import shlex
import subprocess

from ..args import arg
from ..command import command
from ..result import Result
from ..util import abs_path, flatten_args, printer, StreamOptions


@command
def local(args: arg(container=list),
          background=False,
          cd=None,
          environ: arg(type=dict) = None,
          replace_env=False,
          paths=(),
          shell: arg(type=bool) = None,
          stdout: arg(type=StreamOptions) = None,
          stderr: arg(type=StreamOptions) = None,
          echo=False,
          raise_on_error=True,
          dry_run=False,
          ) -> Result:
    """Run a local command via :func:`subprocess.run`.

    Args:
        args (list|str): A list of args or a shell command.
        background (bool): Run process in background? If this is set,
            the command will be run in the background via
            :class:`subprocess.Popen` then this function will
            immediately return. The call site will need to wait on the
            returned :class:`Popen` object using :meth:`Popen.wait` or
            by some other means (perhaps by starting another
            long-running process in the foreground).
        cd (str): Working directory to change to first.
        environ (dict): Additional environment variables to pass to the
            subprocess.
        replace_env (bool): If set, only pass env variables from
            ``environ`` to the subprocess.
        paths (list): A list of additional paths.
        shell (bool): Run as a shell command? The default is to run in
            shell mode if ``args`` is a string. This flag can be used to
            force a list of args to be run as a shell command too.
        stdout (StreamOptions): What to do with stdout (capture, hide,
            or show).
        stderr (StreamOptions): Same as ``stdout``.
        echo (bool): Whether to echo the command before running it.
        raise_on_error (bool): Whether to raise an exception when the
            subprocess returns a non-zero exit code.
        dry_run (bool): If set, print command instead of running it.

    Returns:
        - :class:`Result`: When the command is run in the foreground.
        - :class:`subprocess.Popen`: When the command is run in the
            background.

    Raises:
        Result: When the subprocess returns a non-zero exit code (and
            ``raise_on_error`` is set).

    """
    if isinstance(args, str):
        if shell is None:
            shell = True
    else:
        args = flatten_args(args, join=shell)

    if cd:
        cd = abs_path(cd)
        cd_passed = True
    else:
        cd_passed = False

    environ = {k: str(v) for k, v in (environ or {}).items()}

    if replace_env:
        subprocess_env = environ.copy()
    else:
        subprocess_env = os.environ.copy()
        if environ:
            subprocess_env.update(environ)

    if paths:
        paths = [paths] if isinstance(paths, str) else paths
        paths = [abs_path(p) for p in paths]
        current_path = subprocess_env.get('PATH')
        if current_path:
            paths.append(current_path)
        path = ':'.join(paths)
        subprocess_env['PATH'] = path

    if stdout:
        stdout = StreamOptions[stdout] if isinstance(stdout, str) else stdout
        stdout = stdout.option

    if stderr:
        stderr = StreamOptions[stderr] if isinstance(stderr, str) else stderr
        stderr = stderr.option

    kwargs = {
        'cwd': cd,
        'env': subprocess_env,
        'shell': shell,
        'stdout': stdout,
        'stderr': stderr,
        'universal_newlines': True,
    }

    display_str = args if shell else ' '.join(shlex.quote(a) for a in args)

    if echo:
        if cd_passed:
            printer.echo('{cd}>'.format_map(locals()), end=' ')
        if not dry_run:
            printer.echo(display_str)

    if dry_run:
        printer.echo('[DRY RUN]', display_str)
        result = Result(args, 0, None, None)
    elif background:
        return subprocess.Popen(args, **kwargs)
    else:
        result = subprocess.run(args, **kwargs)
        result = Result.from_subprocess_result(result)

    if result.return_code and raise_on_error:
        raise result

    return result



import datetime
import pathlib
import re
from collections import namedtuple

from ..args import arg
from ..command import command
from ..util import abort, printer, confirm, prompt
from .local import local


@command
def release(version: 'Version to release' = None,
            version_file: arg(type=pathlib.Path, help='File __version__ is in') = None,
            date: 'Release data' = None,
            tag_name: 'Release tag (defaults to version)' = None,
            next_version: 'Anticipated version of next release' = None,
            prepare: 'Run release preparation tasks?' = True,
            merge: 'Run merge tasks' = True,
            target_branch: 'Branch to merge into' = 'master',
            create_tag: 'Create tag when merging?' = True,
            resume: 'Run resume development tasks?' = True,
            test: 'Run tests first?' = True,
            test_command: 'Test command' = 'tox',
            yes: 'Run without being prompted for any confirmations' = False):
    """Make a release.

    Tries to guess the release version based on the current version and
    the next version based on the release version.

    Steps:
        - Prepare release:
            - Update ``__version__`` in version file (typically
              ``package/__init__.py`` or ``src/package/__init__.py``)
            - Update next version header in change log
            - Commit version file and change log with prepare message
        - Merge to target branch (master by default):
            - Merge current branch into target branch with merge message
        - Create tag:
            - Add annotated tag for latest version; when merging, the
              tag will point at the merge commit on the target branch;
              when not merging, the tag will point at the prepare
              release commit on the current branch
        - Resume development:
            - Update version in version file to next version
            - Add in-progress section for next version to change log
            - Commit version file and change log with resume message

    Caveats:
        - Releases cannot be made from the target branch; some other
          branch must be checked out before running the ``release``
          command
        - The next version will have the dev marker '.dev0' appended to
          it
        - Change log must be in Markdown format; release section headers
          must be second-level (i.e., start with ##)
        - Change log must be named CHANGELOG or CHANGELOG.md
        - The first release section header in the change log will be
          updated, so there always needs to be an in-progress section
          for the next version
        - Building distributions and uploading to PyPI isn't handled;
          you'll need to run `python setup.py sdist` (for example) and
          upload the result with `twine upload` (or by some other means)

    """
    current_branch = get_current_branch()
    if current_branch == target_branch:
        abort(1, 'Cannot release from {target_branch} branch'.format_map(locals()))

    print_step('Preparing?', prepare)
    print_step('Merging?', merge)
    print_step('Tagging?', create_tag)
    print_step('Resuming development?', resume)
    print_step('Testing?', test)

    if not version_file:
        version_file, version_line_number, current_version = find_version_file()
    else:
        version_line_number, current_version = get_current_version(version_file)

    if not version:
        if current_version:
            version = current_version
        else:
            message = (
                'Current version not set in version file, so release version needs to be passed '
                'explicitly')
            abort(2, message)

    if not tag_name:
        tag_name = version

    date = date or datetime.date.today().isoformat()

    if not next_version:
        next_version = get_next_version(version)

    change_log = find_change_log()
    change_log_line_number = find_change_log_section(change_log, version)

    info = ReleaseInfo(
        current_branch,
        target_branch,
        version_file,
        version_line_number,
        version,
        tag_name,
        date,
        next_version,
        change_log,
        change_log_line_number,
        not yes,
    )

    printer.info('Version:', info.version)
    printer.info('Tag name:', info.tag_name)
    printer.info('Release date:', info.date)
    printer.info('Next version:', info.next_version)

    if info.confirmation_required:
        msg = 'Continue with release?: {info.version} - {info.date}'.format(info=info)
        confirm(msg, abort_on_unconfirmed=True)
    else:
        printer.warning('Continuing with release: {info.version} - {info.date}')

    if test:
        print_step_header('Testing')
        local(test_command)
    else:
        printer.warning('Skipping tests')

    if prepare:
        prepare_release(info)

    if merge:
        merge_to_target_branch(info)

    if create_tag:
        create_release_tag(info, merge)

    if resume:
        resume_development(info)


ReleaseInfo = namedtuple('ReleaseInfo', (
    'current_branch',
    'target_branch',
    'version_file',
    'version_line_number',
    'version',
    'tag_name',
    'date',
    'next_version',
    'change_log',
    'change_log_line_number',
    'confirmation_required',
))


# Steps


def prepare_release(info):
    print_step_header('Preparing release', info.version, 'on', info.date)

    updated_init_line = "__version__ = '{info.version}'\n".format_map(locals())
    updated_change_log_line = '## {info.version} - {info.date}\n'.format_map(locals())

    update_line(info.version_file, info.version_line_number, updated_init_line)
    update_line(info.change_log, info.change_log_line_number, updated_change_log_line)

    local(('git', 'diff', info.version_file, info.change_log))

    if info.confirmation_required:
        confirm('Commit these changes?', abort_on_unconfirmed=True)
    else:
        printer.warning('Committing changes')

    msg = 'Prepare release {info.version}'.format_map(locals())
    msg = prompt('Commit message', default=msg)
    local(('git', 'commit', info.version_file, info.change_log, '-m', msg))


def merge_to_target_branch(info):
    print_step_header(
        'Merging', info.current_branch, 'into', info.target_branch, 'for release', info.version)

    local(('git', 'log', '--oneline', '--reverse', '{info.target_branch}..'.format_map(locals())))

    if info.confirmation_required:
        msg = (
            'Merge these changes from {info.current_branch} '
            'into {info.target_branch} '
            'for release {info.version}?')
        msg = msg.format_map(locals())
        confirm(msg, abort_on_unconfirmed=True)
    else:
        printer.warning(
            'Merging changes from', info.current_branch,
            'into', info.target_branch,
            'for release', info.release)

    local(('git', 'checkout', info.target_branch))

    msg = 'Merge branch \'{info.current_branch}\' for release {info.version}'
    msg = msg.format_map(locals())
    msg = prompt('Commit message', default=msg)
    local(('git', 'merge', '--no-ff', info.current_branch, '-m', msg))

    local(('git', 'checkout', info.current_branch))


def create_release_tag(info, merge):
    print_step_header('Tagging release', info.version)

    if merge:
        local(('git', 'checkout', info.target_branch))

    local('git log -1 --oneline')

    if info.confirmation_required:
        confirmed = confirm('Tag this commit?')
    else:
        printer.warning('Tagging commit')
        confirmed = True

    if confirmed:
        msg = 'Release {info.version}'.format_map(locals())
        local(('git', 'tag', '-a', '-m', msg, info.tag_name))

    if merge:
        local(('git', 'checkout', info.current_branch))

    if not confirmed:
        abort()


def resume_development(info):
    print_step_header('Resuming development at', info.next_version)

    updated_init_line = "__version__ = '{info.next_version}.dev0'\n".format_map(locals())
    new_change_log_lines = [
        '## {info.next_version} - unreleased\n\n'.format_map(locals()),
        'In progress...\n\n',
    ]

    update_line(info.version_file, info.version_line_number, updated_init_line)

    with info.change_log.open() as fp:
        lines = fp.readlines()

    lines = (
        lines[:info.change_log_line_number] +
        new_change_log_lines +
        lines[info.change_log_line_number:]
    )

    with info.change_log.open('w') as fp:
        fp.writelines(lines)

    local(('git', 'diff', info.version_file, info.change_log))

    if info.confirmation_required:
        confirm('Commit these changes?', abort_on_unconfirmed=True)
    else:
        printer.warning('Committing changes')

    msg = 'Resume development at {info.next_version}'.format_map(locals())
    msg = prompt('Commit message', default=msg)
    local(('git', 'commit', info.version_file, info.change_log, '-m', msg))


# Utilities


def print_step(message, flag):
    printer.info(message, end=' ', flush=True)
    printer.print('yes' if flag else 'no', color='green' if flag else 'red')


def print_step_header(arg, *args):
    printer.hr('\n%s' % arg, *args, end='\n\n')


def get_current_branch():
    result = local('git rev-parse --abbrev-ref HEAD', stdout='capture')
    return result.stdout.strip()


def get_latest_tag():
    result = local('git rev-list --tags --max-count=1', stdout='capture')
    revision = result.stdout.strip()
    result = local(('git', 'describe', '--tags', revision), stdout='capture')
    tag = result.stdout.strip()
    return tag


def find_version_file():
    # Try to find __version__ in:
    #
    # - package/__init__.py
    # - namespace_package/package/__init__.py
    # - src/package/__init__.py
    # - src/namespace_package/package/__init__.py
    cwd = pathlib.Path.cwd()
    candidates = []
    candidates.extend(cwd.glob('*/__init__.py'))
    candidates.extend(cwd.glob('*/*/__init__.py'))
    candidates.extend(cwd.glob('src/*/__init__.py'))
    candidates.extend(cwd.glob('src/*/*/__init__.py'))
    for candidate in candidates:
        result = get_current_version(candidate, False)
        if result is not None:
            return (candidate,) + result
    candidates = '\n    '.join(str(candidate) for candidate in candidates)
    message = 'Could not find file containing __version__; tried:\n    {candidates}'
    message = message.format_map(locals())
    abort(3, message)


def get_current_version(version_file, abort_on_not_found=True):
    # Extract current version from __version__ in version file.
    #
    # E.g.: __version__ = '1.0.dev0'
    version_re = r"""^__version__ *= *(['"])((?P<version>.+?)(?P<dev_marker>\.dev\d+)?)?\1 *$"""
    with version_file.open() as fp:
        for line_number, line in enumerate(fp):
            match = re.search(version_re, line)
            if match:
                return line_number, match.group('version')
    if abort_on_not_found:
        abort(4, 'Could not find __version__ in {version_file}'.format_map(locals()))


def get_next_version(current_version):
    next_version_re = r'^(?P<major>\d+)\.(?P<minor>\d+)(?P<rest>.*)$'
    match = re.search(next_version_re, current_version)
    if match:
        major = match.group('major')
        minor = match.group('minor')

        major = int(major)
        minor = int(minor)

        rest = match.group('rest')
        patch_re = r'^\.(?P<patch>\d+)$'
        match = re.search(patch_re, rest)

        if match:
            # X.Y.Z
            minor += 1
            patch = match.group('patch')
            next_version = '{major}.{minor}.{patch}'.format_map(locals())
        else:
            pre_re = r'^(?P<pre_marker>a|b|rc)(?P<pre_version>\d+)$'
            match = re.search(pre_re, rest)
            if match:
                # X.YaZ
                pre_marker = match.group('pre_marker')
                pre_version = match.group('pre_version')
                pre_version = int(pre_version)
                pre_version += 1
                next_version = '{major}.{minor}{pre_marker}{pre_version}'.format_map(locals())
            else:
                # X.Y or starts with X.Y (but is not X.Y.Z or X.YaZ)
                minor += 1
                next_version = '{major}.{minor}'.format_map(locals())

        return next_version

    msg = 'Cannot automatically determine next version from {version}'.format_map(locals())
    abort(5, msg)


def find_change_log():
    change_log_candidates = ['CHANGELOG', 'CHANGELOG.md']
    for candidate in change_log_candidates:
        path = pathlib.Path.cwd() / candidate
        if path.is_file():
            return path
    message = 'Could not find change log; tried {candidates}'
    message = message.format(candidates=', '.join(change_log_candidates))
    abort(6, message)


def find_change_log_section(change_log, version):
    # Find the first line that starts with '##'. Extract the version and
    # date from that line. The version must be the specified release
    # version OR the date must be the literal string 'unreleased'.

    # E.g.: ## 1.0.0 - unreleased
    change_log_header_re = r'^## (?P<version>.+) - (?P<date>.+)$'

    with change_log.open() as fp:
        for line_number, line in enumerate(fp):
            match = re.search(change_log_header_re, line)
            if match:
                found_version = match.group('version')
                found_date = match.group('date')
                if found_version == version:
                    if found_date != 'unreleased':
                        printer.warning('Re-releasing', version)
                elif found_date == 'unreleased':
                    if found_version != version:
                        printer.warning('Replacing', found_version, 'with', version)
                else:
                    msg = (
                        'Expected version {version} or release date "unreleased"; got:\n\n'
                        '    {line}'
                    ).format_map(locals())
                    abort(7, msg)
                return line_number

    abort(8, 'Could not find section in change log')


def update_line(path, line_number, content):
    with path.open() as fp:
        lines = fp.readlines()
    lines[line_number] = content
    with path.open('w') as fp:
        fp.writelines(lines)



import glob
import importlib
import os
import shlex
from contextlib import redirect_stderr

from ..args import arg
from ..collection import Collection
from ..command import command
from ..const import DEFAULT_COMMANDS_MODULE
from ..run import run


@command
def complete(
    command_line,
    current_token,
    position,
    shell: arg(choices=('bash', 'fish')),
):
    """Find completions for current command.

    This assumes that we'll handle all completion logic here and that
    the shell's automatic file name completion is disabled.

    Args:
        command_line: Command line
        current_token: Token at cursor
        position: Current cursor position
        shell: Name of shell

    """
    position = int(position)
    tokens = shlex.split(command_line[:position])
    all_argv, run_argv, command_argv = run.partition_argv(tokens[1:])

    with open(os.devnull, 'wb') as devnull_fp:
        with redirect_stderr(devnull_fp):
            run_args = run.parse_args(run_argv)

    module = run_args.get('commands_module')
    module = module or DEFAULT_COMMANDS_MODULE
    module = normalize_path(module)

    try:
        base_collection = Collection.load_from_module(module)
    except Exception:
        base_collection = {}

    return _complete(run, base_collection, command_line, current_token, position, shell)


@command
def complete_base_command(
    base_command,
    command_line,
    current_token,
    position,
    shell: arg(choices=('bash', 'fish')),
):
    """Find completions for a base command and its subcommands.

    This assumes that we'll handle all completion logic here and that
    the shell's automatic file name completion is disabled.

    Args:
        base_command: Path to base command as package.module:command
        command_line: Command line
        current_token: Token at cursor
        position: Current cursor position
        shell: Name of shell

    """
    module_name, base_command_name = base_command.rsplit('.', 1)
    module = importlib.import_module(module_name)
    base_command = getattr(module, base_command_name)
    base_collection = {c.base_name: c for c in base_command.subcommands}
    return _complete(base_command, base_collection, command_line, current_token, position, shell)


def _complete(
    base_command,
    base_collection,
    command_line,
    current_token,
    position,
    shell,
):
    position = int(position)
    tokens = shlex.split(command_line[:position])

    # XXX: This isn't quite correct because it will return the base
    #      command in case where it shouldn't. E.g., if `run xxx` is
    #      typed in, you'll get completions for `run` when you should
    #      probably just get nothing.
    found_command = find_command(base_collection, tokens) or base_command

    if found_command.is_subcommand:
        if found_command.is_base_command:
            collection = {c.base_name: c for c in found_command.subcommands}
        else:
            collection = {}
    else:
        collection = base_collection

    if current_token:
        # Completing either a command name, option name, or path.
        if current_token.startswith('-'):
            if current_token not in found_command.option_map:
                print_command_options(found_command, current_token)
        else:
            print_commands(collection, shell)
            path = os.path.expanduser(current_token)
            path = os.path.expandvars(path)
            paths = glob.glob('%s*' % path)
            if paths:
                for entry in paths:
                    if os.path.isdir(entry):
                        print('%s/' % entry)
                    else:
                        print(entry)
    else:
        # Completing option value. If a value isn't expected, show the
        # options for the current command and the list of commands
        # instead.
        option = found_command.option_map.get(tokens[-1])

        if option and option.takes_value:
            if option.choices:
                for choice in option.choices:
                    print(choice)
            else:
                for entry in os.listdir():
                    if os.path.isdir(entry):
                        print('%s/' % entry)
                    else:
                        print(entry)
        else:
            print_command_options(found_command)
            print_commands(collection, shell)


def normalize_path(path):
    if not path:
        return path
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    path = os.path.abspath(path)
    return path


def find_command(collection, tokens):
    for token in reversed(tokens):
        if token in collection:
            return collection[token]


def print_commands(collection, shell):
    for name in collection:
        cmd = collection[name]
        description = cmd.description.splitlines()[0].strip() if cmd.description else ''
        if shell in ('sh', 'bash'):
            print(name)
        elif shell == 'fish':
            print('{name}\t{description}'.format_map(locals()))


def print_command_options(cmd, prefix=''):
    for name, cmd_arg in cmd.args.items():
        for option in cmd_arg.all_options:
            if option.startswith(prefix):
                print(option)



import re


def camel_to_underscore(name):
    """Convert camel case name to underscore name.

    Examples::

        >>> camel_to_underscore('HttpRequest')
        'http_request'
        >>> camel_to_underscore('httpRequest')
        'http_request'
        >>> camel_to_underscore('HTTPRequest')
        'http_request'
        >>> camel_to_underscore('myHTTPRequest')
        'my_http_request'
        >>> camel_to_underscore('MyHTTPRequest')
        'my_http_request'
        >>> camel_to_underscore('my_http_request')
        'my_http_request'
        >>> camel_to_underscore('MyHTTPRequestXYZ')
        'my_http_request_xyz'
        >>> camel_to_underscore('_HTTPRequest')
        '_http_request'
        >>> camel_to_underscore('Request')
        'request'
        >>> camel_to_underscore('REQUEST')
        'request'
        >>> camel_to_underscore('_Request')
        '_request'
        >>> camel_to_underscore('__Request')
        '__request'
        >>> camel_to_underscore('_request')
        '_request'
        >>> camel_to_underscore('Request_')
        'request_'

    """
    name = re.sub(r'(?<!\b)(?<!_)([A-Z][a-z])', r'_\1', name)
    name = re.sub(r'(?<!\b)(?<!_)([a-z])([A-Z])', r'\1_\2', name)
    name = name.lower()
    return name



import functools
import importlib
import os

from ..exc import RunAborted


def abort(return_code=0, message='Aborted', color=True):
    from .printer import printer
    if message:
        if color is True:
            color = 'error' if return_code else 'warning'
        if color:
            message = printer.colorize(message, color=color)
    raise RunAborted(return_code, message)


def flatten_args(args: list, join=False, *, empty=(None, [], (), '')) -> list:
    """Flatten args and remove empty items.

    Args:
        args: A list of items (typically but not necessarily strings),
            which may contain sub-lists, that will be flattened into
            a single list with empty items removed. Empty items include
            ``None`` and empty lists, tuples, and strings.
        join: If ``True`` or a string, the final flattened list will be
            joined into a single string. The default join string is
            a space.
        empty: Items that are considered empty.

    Returns:
        list|str: The list of args flattened with empty items removed
            and the remaining items converted to strings. If ``join`` is
            specified, the list of flattened args will be joined into
            a single string.

    Examples::

        >>> flatten_args([])
        []
        >>> flatten_args(())
        []
        >>> flatten_args([(), (), [(), ()]])
        []
        >>> flatten_args(['executable', '--flag' if True else None, ('--option', 'value'), [None]])
        ['executable', '--flag', '--option', 'value']
        >>> flatten_args(['executable', '--option', 0])
        ['executable', '--option', '0']

    """
    flat_args = []
    non_empty_args = (arg for arg in args if arg not in empty)
    for arg in non_empty_args:
        if isinstance(arg, (list, tuple)):
            flat_args.extend(flatten_args(arg))
        else:
            flat_args.append(str(arg))
    if join:
        join = ' ' if join is True else join
        flat_args = join.join(flat_args)
    return flat_args


def format_if(value, format_kwargs):
    """Apply format args to value if value or return value as is."""
    if not value:
        return value
    return value.format_map(format_kwargs)


def isatty(stream):
    try:
        return stream.isatty()
    except Exception:
        pass

    try:
        fileno = stream.fileno()
    except Exception:
        pass
    else:
        return os.isatty(fileno)

    return False


def load_object(obj) -> object:
    """Load an object.

    Args:
        obj (str|object): Load the indicated object if this is a string;
            otherwise, return the object as is.

            To load a module, pass a dotted path like 'package.module';
            to load an an object from a module pass a path like
            'package.module:name'.

    Returns:
        object

    """
    if isinstance(obj, str):
        if ':' in obj:
            module_name, obj_name = obj.split(':')
            if not module_name:
                module_name = '.'
        else:
            module_name = obj
        obj = importlib.import_module(module_name)
        if obj_name:
            attrs = obj_name.split('.')
            for attr in attrs:
                obj = getattr(obj, attr)
    return obj


def merge_dicts(*dicts):
    """Merge all dicts.

    Dicts later in the list take precedence over dicts earlier in the
    list.

    """
    return functools.reduce(_merge_dicts, dicts, {})


def _merge_dicts(a, b):
    # Merge dict b into dict a
    a = a.copy()
    if not (isinstance(a, dict) and isinstance(b, dict)):
        raise TypeError(
            'Expected two dicts; got {a.__class__} and {b.__class__}'
            .format_map(locals()))
    for k, v in b.items():
        if k in a and isinstance(a[k], dict):
            v = merge_dicts(a[k], v)
        a[k] = v
    return a



from ..args import arg
from ..command import command
from ..result import Result
from ..util import abs_path, StreamOptions
from .local import local


@command
def sync(source,
         destination,
         host,
         user=None,
         sudo=False,
         run_as=None,
         options=('-rltvz', '--no-perms', '--no-group'),
         excludes=(),
         exclude_from=None,
         delete=False,
         dry_run=False,
         mode='u=rwX,g=rwX,o=',
         quiet=True,
         pull=False,
         # Args passed through to local command:
         stdout: arg(type=StreamOptions) = None,
         stderr: arg(type=StreamOptions) = None,
         echo=False,
         raise_on_error=True,
         ) -> Result:
    """Sync files using rsync.

    By default, a local ``source`` is pushed to a remote
    ``destination``. To pull from a remote ``source`` to a local
    ``destination`` instead, pass ``pull=True``.

    """
    source = abs_path(source, keep_slash=True)
    destination = abs_path(destination, keep_slash=True)
    connection_str = '{user}@{host}'.format_map(locals()) if user else host
    push = not pull

    if sudo:
        rsync_path = ('--rsync-path', 'sudo rsync')
    elif run_as:
        rsync_path = ('--rsync-path', 'sudo -u {run_as} rsync'.format_map(locals()))
    else:
        rsync_path = None

    if push:
        destination = '{connection_str}:{destination}'.format_map(locals())
    else:
        source = '{connection_str}:{source}'.format_map(locals())

    args = (
        'rsync',
        rsync_path,
        options,
        ('--chmod', mode) if mode else None,
        tuple(('--exclude', exclude) for exclude in excludes),
        ('--exclude-from', exclude_from) if exclude_from else None,
        '--delete' if delete else None,
        '--dry-run' if dry_run else None,
        '--quiet' if quiet else None,
        source,
        destination,
    )
    return local(args, stdout=stdout, stderr=stderr, echo=echo, raise_on_error=raise_on_error)



import os
from unittest import TestCase

from .. import arg, command, subcommand
from ..commands import local


class Result:

    def __init__(self, return_code):
        self.return_code = return_code


class TestLocalCommand(TestCase):

    def test_local_ls(self):
        result = local(['ls', '-1'], cd=os.path.dirname(__file__), stdout='capture')
        self.assertIn('__init__.py', result.stdout_lines)
        self.assertTrue(result)


@command
def base(subcommand: arg(default=None)):
    return Result('base {subcommand}'.format_map(locals()))


@subcommand(base)
def sub(subcommand: arg(default=None), optional=None):
    return Result('sub {subcommand} {optional}'.format_map(locals()))


@subcommand(sub)
def subsub(positional, optional=None):
    return Result('subsub {positional} {optional}'.format_map(locals()))


class TestSubcommand(TestCase):

    def test_base_command_subcommand_choices(self):
        arg = base.args['subcommand']
        self.assertIsNotNone(arg.choices)
        self.assertEqual(arg.choices, ['sub'])

    def test_call_base_command(self):
        result = base.console_script(argv=[])
        self.assertEqual(result, 'base None')

    def test_call_subcommand(self):
        result = base.console_script(argv=['sub'])
        self.assertEqual(result, 'sub None None')

    def test_call_subcommand_with_optional(self):
        result = base.console_script(argv=['sub', '--optional', 'b'])
        self.assertEqual(result, 'sub None b')

    def test_call_subsubcommand(self):
        result = base.console_script(argv=['sub', 'subsub', 'a'])
        self.assertEqual(result, 'subsub a None')

    def test_call_subsubcommand_with_optional(self):
        result = base.console_script(argv=['sub', 'subsub', 'a', '--optional', 'b'])
        self.assertEqual(result, 'subsub a b')

    def test_call_subsubcommand_with_shared_args(self):
        @command
        def base1(cmd, a=None):
            return Result('base1({cmd}, {a})'.format_map(locals()))

        @subcommand(base1)
        def sub1(cmd: arg(default=None), a=None, flag=True):
            return Result('sub1({cmd}, {a}, {flag})'.format_map(locals()))

        @sub1.subcommand
        def subsub1(a=None, flag=True):
            return Result('subsub1({a}, {flag})'.format_map(locals()))

        result = base1.console_script(argv=['-a', 'A', 'sub1', '--no-flag'])
        self.assertEqual(result, 'sub1(None, A, False)')

        result = base1.console_script(argv=['-a', 'A', 'sub1', '--no-flag', 'subsub1'])
        self.assertEqual(result, 'subsub1(A, False)')



import os
from setuptools import find_packages, setup


def as_bool(value):
    # Convert env var to bool
    if not isinstance(value, str):
        return value
    return value.lower() in ('1', 'true', 'yes')


with open('runcommands/__init__.py') as fp:
    for line in fp:
        if line.startswith('__version__'):
            __version__ = line.split('=')[1].strip()[1:-1]

with open('README.rst') as fp:
    long_description = fp.read().strip()

# Install run, runcommand, and runcommands main console scripts by default
default_aliases = 'run runcommand runcommands'
script_names = os.getenv('RUNCOMMANDS_CONSOLE_SCRIPTS', default_aliases).split()
script_path = 'runcommands.__main__:main'
console_scripts = ['{name} = {path}'.format(name=name, path=script_path) for name in script_names]

# Install runcommands-complete completion console script by default
install_complete_console_script = os.getenv('RUNCOMMANDS_INSTALL_COMPLETION_CONSOLE_SCRIPTS', True)
if as_bool(install_complete_console_script):
    console_scripts.extend((
        'runcommands-complete = runcommands.completion:complete.console_script',
        'runcommands-complete-base-command = '
        'runcommands.completion:complete_base_command.console_script',
    ))

# The release console script is *not* installed by default
install_release_console_script = os.getenv('RUNCOMMANDS_INSTALL_RELEASE_CONSOLE_SCRIPT', False)
if as_bool(install_release_console_script):
    console_scripts.append('make-release = runcommands.commands:release.console_script')

setup(
    name='runcommands',
    version=__version__,
    license='MIT',
    author='Wyatt Baldwin',
    author_email='self@wyattbaldwin.com',
    description='A simple command runner',
    long_description=long_description,
    keywords=['run', 'commands', 'console', 'scripts', 'terminal'],
    url='https://github.com/wylee/runcommands',
    python_requires='>=3.5',
    install_requires=[
        'Jinja2>=2.10',
        'PyYAML>=5.1',
    ],
    packages=find_packages(),
    include_package_data=True,
    extras_require={
        'dev': [
            'coverage',
            'flake8',
            'Sphinx',
            'tox',
            'twine',
        ],
    },
    entry_points={
        'console_scripts': console_scripts,
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



class Data:

    """A bucket for arbitrary data.

    Data can be added and retrieved as attributes (dot notation) or
    items (bracket notation).

    When a ``dict`` is added, it will be converted to an instance of
    :class:`Data`.

    """

    def __init__(self, **data):
        super().__setattr__('__data', {})
        for name, value in data.items():
            self[name] = value

    def __getattr__(self, name):
        data = super().__getattribute__('__data')
        return data[name]

    __getitem__ = __getattr__

    def __setattr__(self, name, value):
        data = super().__getattribute__('__data')
        if isinstance(value, dict):
            value = self.__class__(**value)
        data[name] = value

    __setitem__ = __setattr__
